----- MODULE SCounter -----
EXTENDS Naturals, TLC, Sequences
CONSTANT N, K

(* The algorithm implements a strong counter using CAS objects and atomic registers.
   It uses the universal construction (see the lectures). *)

(* --algorithm SCounterAlg
   
	(* 1. Shared objects *)

		variables R = [ i \in 1..N |-> [ flag |-> FALSE ] ],
			  CS = [ i \in 1..N*K |-> << >> ],

		(* for testing correctness *)
			  time = 0,

	(* 2. Process-local variables *)

			 round = [ i \in 1..N |-> 0 ],
			 counter = [ i \in 1..N |-> 0 ],
			 processed = [ i \in 1..N |-> {} ],
			 incRetVal = [ i \in 1..N |-> 0 ],
			 history = [i \in 1..N |-> << >> ];

	(* 3. The definition of CAS *)

	macro CAS(CS, oldVal, newVal, retVal)
	begin
		if CS = oldVal then CS := newVal; end if;
		retVal := CS;
	end macro

	(* 4. An implementation of a strong counter out of CAS objects and atomic registers *)

	procedure inc()
	variables operations, p, o, req, myreqid, currop, finished = FALSE;
	begin
		getid:	myreqid := round[self];
		prop:	R[self] := [ flag |-> TRUE, reqid |-> myreqid ];
		try:	while ~finished do
		init:		round[self] := round[self] + 1;
				p := 1; 
				o := 1;
				operations := << >>;
		ops1:		while p <= N do
		ops2:			req := R[p];
					if req.flag then
						currop :=  [ pid |-> p, reqid |-> req.reqid ];
						if currop \notin processed[self] then 
							operations := Append(operations, currop);
						end if;
					end if;
					p := p + 1;
				end while;
		decide:		CAS(CS[round[self]], << >>, operations, operations);
		exec1:		while o <= Len(operations) do
		exec2:			currop := operations[o];
					counter[self] := counter[self] + 1;
					if (currop.pid = self) /\ (currop.reqid = myreqid) then 
						incRetVal[self] := counter[self];
						finished := TRUE;
					end if;
					processed[self] := processed[self] \cup { currop };
					o := o + 1;
				end while;
			end while;
		reset:	R[self] := [ flag |-> FALSE ];
			return;
	end procedure

	(* 5. Test procedure *)

	process Proc \in 1..N
	variables tsbeg, tsend, s = 1;
	begin
		test1:	while s <= K do
		ts1:		tsbeg := time;
				time := time + 1;
		doinc:		call inc();
		ts2:		tsend := time;
				time := time + 1;
		save:		history[self] := Append(history[self], [ tsb |-> tsbeg, tse |-> tsend, val |-> incRetVal[self] ]);
				s := s + 1;
			end while
	end process

   end algorithm
*)

\* After this line you put the definitions TLA+ operators used by +CAL the algorithm

\* BEGIN TRANSLATION
VARIABLES R, CS, time, round, counter, processed, incRetVal, history, pc, 
          stack, operations, p, o, req, myreqid, currop, finished, tsbeg, 
          tsend, s

vars == << R, CS, time, round, counter, processed, incRetVal, history, pc, 
           stack, operations, p, o, req, myreqid, currop, finished, tsbeg, 
           tsend, s >>

ProcSet == (1..N)

Init == (* Global variables *)
        /\ R = [ i \in 1..N |-> [ flag |-> FALSE ] ]
        /\ CS = [ i \in 1..N*K |-> << >> ]
        /\ time = 0
        /\ round = [ i \in 1..N |-> 0 ]
        /\ counter = [ i \in 1..N |-> 0 ]
        /\ processed = [ i \in 1..N |-> {} ]
        /\ incRetVal = [ i \in 1..N |-> 0 ]
        /\ history = [i \in 1..N |-> << >> ]
        (* Procedure inc *)
        /\ operations = [ self \in ProcSet |-> {}]
        /\ p = [ self \in ProcSet |-> {}]
        /\ o = [ self \in ProcSet |-> {}]
        /\ req = [ self \in ProcSet |-> {}]
        /\ myreqid = [ self \in ProcSet |-> {}]
        /\ currop = [ self \in ProcSet |-> {}]
        /\ finished = [ self \in ProcSet |-> FALSE]
        (* Process Proc *)
        /\ tsbeg = [self \in 1..N |-> {}]
        /\ tsend = [self \in 1..N |-> {}]
        /\ s = [self \in 1..N |-> 1]
        /\ stack = [self \in ProcSet |-> << >>]
        /\ pc = [self \in ProcSet |-> CASE self \in 1..N -> "test1"]

getid(self) == /\ pc[self] = "getid"
               /\ myreqid' = [myreqid EXCEPT ![self] = round[self]]
               /\ pc' = [pc EXCEPT ![self] = "prop"]
               /\ UNCHANGED << R, CS, time, round, counter, processed, 
                               incRetVal, history, stack, operations, p, o, 
                               req, currop, finished, tsbeg, tsend, s >>

prop(self) == /\ pc[self] = "prop"
              /\ R' = [R EXCEPT ![self] = [ flag |-> TRUE, reqid |-> myreqid[self] ]]
              /\ pc' = [pc EXCEPT ![self] = "try"]
              /\ UNCHANGED << CS, time, round, counter, processed, incRetVal, 
                              history, stack, operations, p, o, req, myreqid, 
                              currop, finished, tsbeg, tsend, s >>

try(self) == /\ pc[self] = "try"
             /\ IF ~finished[self]
                   THEN /\ pc' = [pc EXCEPT ![self] = "init"]
                   ELSE /\ pc' = [pc EXCEPT ![self] = "reset"]
             /\ UNCHANGED << R, CS, time, round, counter, processed, incRetVal, 
                             history, stack, operations, p, o, req, myreqid, 
                             currop, finished, tsbeg, tsend, s >>

init(self) == /\ pc[self] = "init"
              /\ round' = [round EXCEPT ![self] = round[self] + 1]
              /\ p' = [p EXCEPT ![self] = 1]
              /\ o' = [o EXCEPT ![self] = 1]
              /\ operations' = [operations EXCEPT 
                                  ![self] = << >>]
              /\ pc' = [pc EXCEPT ![self] = "ops1"]
              /\ UNCHANGED << R, CS, time, counter, processed, incRetVal, 
                              history, stack, req, myreqid, currop, finished, 
                              tsbeg, tsend, s >>

ops1(self) == /\ pc[self] = "ops1"
              /\ IF p[self] <= N
                    THEN /\ pc' = [pc EXCEPT ![self] = "ops2"]
                    ELSE /\ pc' = [pc EXCEPT ![self] = "decide"]
              /\ UNCHANGED << R, CS, time, round, counter, processed, 
                              incRetVal, history, stack, operations, p, o, req, 
                              myreqid, currop, finished, tsbeg, tsend, s >>

ops2(self) == /\ pc[self] = "ops2"
              /\ req' = [req EXCEPT ![self] = R[p[self]]]
              /\ IF req'[self].flag
                    THEN /\ currop' = [currop EXCEPT 
                                         ![self] = [ pid |-> p[self], reqid |-> req'[self].reqid ]]
                         /\ IF currop'[self] \notin processed[self]
                               THEN /\ operations' = [operations EXCEPT 
                                                        ![self] = Append(operations[self], currop'[self])]
                               ELSE /\ TRUE
                                    /\ UNCHANGED operations
                    ELSE /\ TRUE
                         /\ UNCHANGED << operations, currop >>
              /\ p' = [p EXCEPT ![self] = p[self] + 1]
              /\ pc' = [pc EXCEPT ![self] = "ops1"]
              /\ UNCHANGED << R, CS, time, round, counter, processed, 
                              incRetVal, history, stack, o, myreqid, finished, 
                              tsbeg, tsend, s >>

decide(self) == /\ pc[self] = "decide"
                /\ IF (CS[round[self]]) = (<< >>)
                      THEN /\ CS' = [CS EXCEPT ![round[self]] = operations[self]]
                      ELSE /\ TRUE
                           /\ UNCHANGED CS
                /\ operations' = [operations EXCEPT 
                                    ![self] = CS'[round[self]]]
                /\ pc' = [pc EXCEPT ![self] = "exec1"]
                /\ UNCHANGED << R, time, round, counter, processed, incRetVal, 
                                history, stack, p, o, req, myreqid, currop, 
                                finished, tsbeg, tsend, s >>

exec1(self) == /\ pc[self] = "exec1"
               /\ IF o[self] <= Len(operations[self])
                     THEN /\ pc' = [pc EXCEPT 
                                      ![self] = "exec2"]
                     ELSE /\ pc' = [pc EXCEPT 
                                      ![self] = "try"]
               /\ UNCHANGED << R, CS, time, round, counter, processed, 
                               incRetVal, history, stack, operations, p, o, 
                               req, myreqid, currop, finished, tsbeg, tsend, s >>

exec2(self) == /\ pc[self] = "exec2"
               /\ currop' = [currop EXCEPT ![self] = operations[self][o[self]]]
               /\ counter' = [counter EXCEPT ![self] = counter[self] + 1]
               /\ IF (currop'[self].pid = self) /\ (currop'[self].reqid = myreqid[self])
                     THEN /\ incRetVal' = [incRetVal EXCEPT ![self] = counter'[self]]
                          /\ finished' = [finished EXCEPT 
                                            ![self] = TRUE]
                     ELSE /\ TRUE
                          /\ UNCHANGED << incRetVal, finished >>
               /\ processed' = [processed EXCEPT ![self] = processed[self] \cup { currop'[self] }]
               /\ o' = [o EXCEPT ![self] = o[self] + 1]
               /\ pc' = [pc EXCEPT ![self] = "exec1"]
               /\ UNCHANGED << R, CS, time, round, history, stack, operations, 
                               p, req, myreqid, tsbeg, tsend, s >>

reset(self) == /\ pc[self] = "reset"
               /\ R' = [R EXCEPT ![self] = [ flag |-> FALSE ]]
               /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
               /\ operations' = [operations EXCEPT 
                                   ![self] = Head(stack[self]).operations]
               /\ p' = [p EXCEPT ![self] = Head(stack[self]).p]
               /\ o' = [o EXCEPT ![self] = Head(stack[self]).o]
               /\ req' = [req EXCEPT ![self] = Head(stack[self]).req]
               /\ myreqid' = [myreqid EXCEPT ![self] = Head(stack[self]).myreqid]
               /\ currop' = [currop EXCEPT ![self] = Head(stack[self]).currop]
               /\ finished' = [finished EXCEPT 
                                 ![self] = Head(stack[self]).finished]
               /\ stack' = [stack EXCEPT ![self] = Tail(stack[self])]
               /\ UNCHANGED << CS, time, round, counter, processed, incRetVal, 
                               history, tsbeg, tsend, s >>

inc(self) == getid(self) \/ prop(self) \/ try(self) \/ init(self)
                \/ ops1(self) \/ ops2(self) \/ decide(self) \/ exec1(self)
                \/ exec2(self) \/ reset(self)

test1(self) == /\ pc[self] = "test1"
               /\ IF s[self] <= K
                     THEN /\ pc' = [pc EXCEPT 
                                      ![self] = "ts1"]
                     ELSE /\ pc' = [pc EXCEPT 
                                      ![self] = "Done"]
               /\ UNCHANGED << R, CS, time, round, counter, processed, 
                               incRetVal, history, stack, operations, p, o, 
                               req, myreqid, currop, finished, tsbeg, tsend, s >>

ts1(self) == /\ pc[self] = "ts1"
             /\ tsbeg' = [tsbeg EXCEPT ![self] = time]
             /\ time' = time + 1
             /\ pc' = [pc EXCEPT ![self] = "doinc"]
             /\ UNCHANGED << R, CS, round, counter, processed, incRetVal, 
                             history, stack, operations, p, o, req, myreqid, 
                             currop, finished, tsend, s >>

doinc(self) == /\ pc[self] = "doinc"
               /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "inc",
                                                        pc        |->  "ts2",
                                                        operations |->  operations[self],
                                                        p         |->  p[self],
                                                        o         |->  o[self],
                                                        req       |->  req[self],
                                                        myreqid   |->  myreqid[self],
                                                        currop    |->  currop[self],
                                                        finished  |->  finished[self] ] >>
                                                    \o stack[self]]
               /\ operations' = [operations EXCEPT 
                                   ![self] = {}]
               /\ p' = [p EXCEPT ![self] = {}]
               /\ o' = [o EXCEPT ![self] = {}]
               /\ req' = [req EXCEPT ![self] = {}]
               /\ myreqid' = [myreqid EXCEPT ![self] = {}]
               /\ currop' = [currop EXCEPT ![self] = {}]
               /\ finished' = [finished EXCEPT 
                                 ![self] = FALSE]
               /\ pc' = [pc EXCEPT ![self] = "getid"]
               /\ UNCHANGED << R, CS, time, round, counter, processed, 
                               incRetVal, history, tsbeg, tsend, s >>

ts2(self) == /\ pc[self] = "ts2"
             /\ tsend' = [tsend EXCEPT ![self] = time]
             /\ time' = time + 1
             /\ pc' = [pc EXCEPT ![self] = "save"]
             /\ UNCHANGED << R, CS, round, counter, processed, incRetVal, 
                             history, stack, operations, p, o, req, myreqid, 
                             currop, finished, tsbeg, s >>

save(self) == /\ pc[self] = "save"
              /\ history' = [history EXCEPT ![self] = Append(history[self], [ tsb |-> tsbeg[self], tse |-> tsend[self], val |-> incRetVal[self] ])]
              /\ s' = [s EXCEPT ![self] = s[self] + 1]
              /\ pc' = [pc EXCEPT ![self] = "test1"]
              /\ UNCHANGED << R, CS, time, round, counter, processed, 
                              incRetVal, stack, operations, p, o, req, myreqid, 
                              currop, finished, tsbeg, tsend >>

Proc(self) == test1(self) \/ ts1(self) \/ doinc(self) \/ ts2(self)
                 \/ save(self)

Next == (\E self \in ProcSet: inc(self))
           \/ (\E self \in 1..N: Proc(self))
           \/ (* Disjunct to prevent deadlock on termination *)
              (\A self \in ProcSet: pc[self] = "Done" /\ UNCHANGED vars)

Spec == Init /\ [][Next]_vars

Termination == <>(\A self \in ProcSet: pc[self] = "Done")

\* END TRANSLATION

\* After this line you put the invariants you want to check against (in TLA+)

(* 6. Invariant that checks for atomicity of the strong counter at the end of the execution *)

SCAtomicity ==  \* if all processes finished, then
		(\A i \in 1..N : pc[i] = "Done") => (
			\* some of them got value 1 from the counter, and
			(\E j \in 1..N : history[j][1].val = 1) /\
			\* no two processes ever got the same value, and
			(\A j \in 1..N, k \in 1..K, l \in 1..N, m \in 1..K : (
				( (j # l) \/ (k # m) ) => (history[j][k].val # history[l][m].val) ) ) /\
			\* if an operation returned a value c > 1, then some other operation,
			\* preceding or concurrent, returned value c-1
			(\A j \in 1..N, k \in 1..K : ( (history[j][k].val > 1) => 
				(\E l \in 1..N, m \in 1..K : (
					(history[l][m].val = history[j][k].val - 1) /\
					(history[l][m].tsb < history[j][k].tse) ) ) ) ) ) 

=====
